'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd
from lxml import etree, objectify

from pm4py.objects.ocel import constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel.util import filtering_utils
from pm4py.util import exec_utils, dt_parsing, pandas_utils
from pm4py.objects.ocel.util import ocel_consistency


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    INTERNAL_INDEX = constants.PARAM_INTERNAL_INDEX
    QUALIFIER = constants.PARAM_QUALIFIER
    CHANGED_FIELD = constants.PARAM_CHNGD_FIELD
    ENCODING = "encoding"


def parse_xml(value, tag_str_lower, parser):
    if "float" in tag_str_lower:
        if value == "null":
            return 0
        return float(value)
    elif "date" in tag_str_lower:
        return embed_date_parser(parser.apply, value)
    return str(value)


def embed_date_parser(date_parser, x):
    try:
        return date_parser(x)
    except BaseException:
        from dateutil.parser import parse

        try:
            return parse(x)
        except BaseException:
            return parse(x, fuzzy=True)


def apply(file_path: str, parameters: Optional[Dict[Any, Any]] = None) -> OCEL:
    if parameters is None:
        parameters = {}

    events_list = []
    relations_list = []
    objects_list = []
    object_changes_list = []
    o2o_list = []

    encoding = exec_utils.get_param_value(
        Parameters.ENCODING, parameters, None
    )

    event_id_column = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, constants.DEFAULT_EVENT_ID
    )
    event_activity_column = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, constants.DEFAULT_EVENT_ACTIVITY
    )
    event_timestamp_column = exec_utils.get_param_value(
        Parameters.EVENT_TIMESTAMP,
        parameters,
        constants.DEFAULT_EVENT_TIMESTAMP,
    )
    object_id_column = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, constants.DEFAULT_OBJECT_ID
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, constants.DEFAULT_OBJECT_TYPE
    )
    internal_index_column = exec_utils.get_param_value(
        Parameters.INTERNAL_INDEX, parameters, constants.DEFAULT_INTERNAL_INDEX
    )
    qualifier_field = exec_utils.get_param_value(
        Parameters.QUALIFIER, parameters, constants.DEFAULT_QUALIFIER
    )
    changed_field = exec_utils.get_param_value(
        Parameters.CHANGED_FIELD, parameters, constants.DEFAULT_CHNGD_FIELD
    )
    obj_type_dict = {}

    date_parser = dt_parsing.parser.get()

    parser = etree.XMLParser(remove_comments=True, encoding=encoding)

    F = open(file_path, "rb")
    tree = objectify.parse(F, parser=parser)
    F.close()

    root = tree.getroot()

    object_type_attributes = {}
    event_type_attributes = {}

    for child in root:
        if child.tag.endswith("object-types"):
            for object_type in child:
                object_type_name = object_type.get("name")
                object_type_attributes[object_type_name] = {}
                for attributes in object_type:
                    for attribute in attributes:
                        attribute_name = attribute.get("name")
                        attribute_type = attribute.get("type")
                        object_type_attributes[object_type_name][
                            attribute_name
                        ] = attribute_type
        elif child.tag.endswith("event-types"):
            for event_type in child:
                event_type_name = event_type.get("name")
                event_type_attributes[event_type_name] = {}
                for attributes in event_type:
                    for attribute in attributes:
                        attribute_name = attribute.get("name")
                        attribute_type = attribute.get("type")
                        event_type_attributes[event_type_name][
                            attribute_name
                        ] = attribute_type
        elif child.tag.endswith("objects"):
            object_id = None
            object_type = None

            for object in child:
                object_id = object.get("id")
                object_type = object.get("type")

                obj_dict = {
                    object_id_column: object_id,
                    object_type_column: object_type,
                }
                obj_type_dict[object_id] = object_type
                obj_attrs = {}

                for child2 in object:
                    if child2.tag.endswith("objects"):
                        for target_object in child2:
                            target_object_id = target_object.get("object-id")
                            qualifier = target_object.get("qualifier")

                            o2o_dict = {
                                object_id_column: object_id,
                                object_id_column + "_2": target_object_id,
                                qualifier_field: qualifier,
                            }
                            o2o_list.append(o2o_dict)

                    elif child2.tag.endswith("attributes"):
                        for attribute in child2:
                            attribute_name = attribute.get("name")
                            attribute_time = attribute.get("time")
                            try:
                                attribute_type = object_type_attributes[
                                    object_type
                                ][attribute_name]
                            except BaseException:
                                attribute_type = "string"
                            attribute_text = parse_xml(
                                attribute.text, attribute_type, date_parser
                            )
                            if (
                                attribute_time == "0"
                                or attribute_time.startswith(
                                    "1970-01-01T00:00:00"
                                )
                            ):
                                time_key = None
                                time_val = attribute_time
                            else:
                                time_val = embed_date_parser(
                                    date_parser.apply, attribute_time
                                )
                                time_key = time_val

                            obj_attrs.setdefault(attribute_name, []).append(
                                (time_key, time_val, attribute_text)
                            )

                for attr_name, entries in obj_attrs.items():
                    base_time_key, base_time_val, base_value = entries[0]
                    obj_dict[attr_name] = base_value
                    for time_key, time_val, value in entries[1:]:
                        obj_change_dict = {
                            object_id_column: object_id,
                            object_type_column: object_type,
                            attr_name: value,
                            changed_field: attr_name,
                            event_timestamp_column: time_val,
                        }
                        object_changes_list.append(obj_change_dict)

                objects_list.append(obj_dict)

        elif child.tag.endswith("events"):
            event_id = None
            event_type = None
            event_time = None

            for event in child:
                event_id = event.get("id")
                event_type = event.get("type")
                event_time = embed_date_parser(
                    date_parser.apply, event.get("time")
                )

                ev_dict = {
                    event_id_column: event_id,
                    event_activity_column: event_type,
                    event_timestamp_column: event_time,
                }

                for child2 in event:
                    if child2.tag.endswith("objects"):
                        for target_object in child2:
                            target_object_id = target_object.get("object-id")
                            qualifier = target_object.get("qualifier")

                            if target_object_id in obj_type_dict:
                                rel_dict = {
                                    event_id_column: event_id,
                                    event_activity_column: event_type,
                                    event_timestamp_column: event_time,
                                    object_id_column: target_object_id,
                                    object_type_column: obj_type_dict[
                                        target_object_id
                                    ],
                                    qualifier_field: qualifier,
                                }
                                relations_list.append(rel_dict)
                    elif child2.tag.endswith("attributes"):
                        for attribute in child2:
                            attribute_name = attribute.get("name")
                            attribute_text = attribute.text
                            try:
                                attribute_type = event_type_attributes[
                                    event_type
                                ][attribute_name]
                            except BaseException:
                                attribute_type = "string"
                            ev_dict[attribute_name] = parse_xml(
                                attribute_text, attribute_type, date_parser
                            )

                events_list.append(ev_dict)

    events_list = (
        pandas_utils.instantiate_dataframe(events_list)
        if events_list
        else None
    )
    objects_list = (
        pandas_utils.instantiate_dataframe(objects_list)
        if objects_list
        else None
    )
    relations_list = (
        pandas_utils.instantiate_dataframe(relations_list)
        if relations_list
        else None
    )
    o2o_list = (
        pandas_utils.instantiate_dataframe(o2o_list) if o2o_list else None
    )
    object_changes_list = (
        pandas_utils.instantiate_dataframe(object_changes_list)
        if object_changes_list
        else None
    )
    globals = {}

    events_list[internal_index_column] = events_list.index
    relations_list[internal_index_column] = relations_list.index

    events_list = events_list.sort_values(
        [event_timestamp_column, internal_index_column]
    )
    relations_list = relations_list.sort_values(
        [event_timestamp_column, internal_index_column]
    )

    del events_list[internal_index_column]
    del relations_list[internal_index_column]

    ocel = OCEL(
        events=events_list,
        objects=objects_list,
        relations=relations_list,
        globals=globals,
        o2o=o2o_list,
        object_changes=object_changes_list,
        parameters=parameters,
    )
    ocel = ocel_consistency.apply(ocel, parameters=parameters)
    ocel = filtering_utils.propagate_relations_filtering(ocel)

    return ocel



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from enum import Enum
from typing import Optional, Dict, Any, Union

import pandas as pd

from pm4py.algo.anonymization.pripel.variants import pripel
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.log.obj import EventLog
from pm4py.util import exec_utils


class Variants(Enum):
    PRIPEL = pripel


DEFAULT_VARIANT = Variants.PRIPEL

VERSIONS = {Variants.PRIPEL}


def apply(log: Union[EventLog, pd.DataFrame], trace_variant_query: Union[EventLog, pd.DataFrame], epsilon: float,
          variant=DEFAULT_VARIANT,
          parameters: Optional[Dict[Any, Any]] = None) -> EventLog:
    """
    PRIPEL (Privacy-preserving event log publishing with contextual information) is a framework to publish event logs
    that fulfill differential privacy. PRIPEL ensures privacy on the level of individual cases instead of the complete
    log. This way, contextual information as well as the long tail process behaviour are preserved, which enables the
    application of a rich set of process analysis techniques.

    PRIPEL is described in:
    Fahrenkrog-Petersen, S.A., van der Aa, H., Weidlich, M. (2020). PRIPEL: Privacy-Preserving Event Log Publishing
    Including Contextual Information. In: Fahland, D., Ghidini, C., Becker, J., Dumas, M. (eds) Business Process
    Management. BPM 2020. Lecture Notes in Computer Science(), vol 12168. Springer, Cham.
    https://doi.org/10.1007/978-3-030-58666-9_7


    Parameters
    -------------
    log
        Event log
    trace_variant_query
        An anonymized trace variant distribution as an EventLog
    epsilon
        Strength of the differential privacy guarantee
    variant
        - Variants.PRIPEL
    parameters
        Parameters of the algorithm, including:
            -Parameters.BLOCKLIST -> Some event logs contain attributes that are equivalent to a case id. For privacy reasons, such attributes must be deleted from the anonymized log. We handle such attributes with this set.
    Returns
    ------------
    anonymised_log
        Anonymised event log
    """

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    trace_variant_query = log_converter.apply(trace_variant_query, variant=log_converter.Variants.TO_EVENT_LOG)

    return exec_utils.get_variant(variant).apply(log, trace_variant_query, epsilon, parameters=parameters)



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import exec_utils, constants, xes_constants
from typing import Optional, Dict, Any, List, Tuple
from pm4py.objects.conversion.log import converter as log_converter
import pandas as pd
from enum import Enum
from pm4py.algo.transformation.to_embeddings.util import similarity
from sentence_transformers import SentenceTransformer


class Parameters(Enum):
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    EMBEDDING_MODEL = "embedding_model"
    ATTRIBUTE_KEY = constants.PARAMETER_CONSTANT_ATTRIBUTE_KEY


def apply(log: pd.DataFrame, parameters: Optional[Dict[Any, Any]] = None) -> Tuple[List[str], List[List[float]]]:
    """
    Computes the embeddings of cases, concatenating all the values of a specified attribute for the events of the case.

    Parameters
    -----------------
    log
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.CASE_ID_KEY => the case identifier column
        - Parameters.EMBEDDING_MODEL => the embedding to be used (default: all-MiniLM-L6-v2)
        - Parameters.ATTRIBUTE_KEY => the attribute to be used for the concatenation

    Returns
    ----------------
    cases_identifiers
        The list of all the case identifiers
    embeddings_list
        The list of embeddings for the considered cases
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)
    attribute_key = exec_utils.get_param_value(Parameters.ATTRIBUTE_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    embedding_model = exec_utils.get_param_value(Parameters.EMBEDDING_MODEL, parameters,
                                                 constants.DEFAULT_EMBEDDING_MODEL)

    log = log_converter.apply(log, variant=log_converter.Variants.TO_DATA_FRAME)

    cases_identifiers = []
    sentences = []

    model = SentenceTransformer(embedding_model)

    cases = log.groupby(case_id_key)[attribute_key].agg(list)
    cases = cases.to_dict()

    for k, v in cases.items():
        cases_identifiers.append(k)
        sentences.append(" ".join(v))

    embeddings_list = list(model.encode(sentences))

    return cases_identifiers, embeddings_list


def keep_top_k_per_similarity(log: pd.DataFrame, target_sentence: str, k: int, cases_identifiers: Optional[List[str]] = None, embeddings_list: Optional[List[List[float]]] = None, parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Keeps the top K cases for (embedding-based) similarity with the given sentence

    Parameters
    ----------------
    log
        Pandas dataframe
    target_sentence
        Target sentence
    k
        Number of similar cases to retain
    cases_identifiers
        (Optional) the list of cases identifiers in the log, as returned by the 'apply' method
    embeddings_list
        (Optional) the list of embeddings for such cases, as returned by the 'apply' method
    parameters
        Other parameters of the method

    Returns
    -----------------
    filtered_log
        Event log filtered on the top K cases according to the similarity metric.
    """
    if parameters is None:
        parameters = {}

    case_id_key = exec_utils.get_param_value(Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME)

    if cases_identifiers is None or embeddings_list is None:
        cases_identifiers, embeddings_list = apply(log, parameters=parameters)

    sim = similarity.apply(target_sentence, embeddings_list)
    cases = [(cases_identifiers[i], sim[i]) for i in range(len(cases_identifiers))]
    cases.sort(key=lambda x: (x[1], x[0]), reverse=True)
    cases = cases[:k]
    cases = [x[0] for x in cases]

    log = log[log[case_id_key].isin(cases)]

    return log



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from itertools import combinations
import networkx as nx
from pm4py.objects.oc_causal_net.utils.filters import filter4
from typing import Tuple, List, Dict
from collections import Counter, defaultdict
from functools import cached_property


class OCCausalNet(object):
    """
    Object-Centric Causal Net capturing dependency graph and marker groups.
    """

    class Marker(object):
        """
        Represents a single marker in an object-centric causal net.
        """

        def __init__(
            self, related_activity, object_type, count_range: Tuple, marker_key: int
        ):
            """
            Constructor

            Parameters
            ----------
            related_activity : str
                Activity that has to fulfill the marker (predecessor or successor)
            object_type : str
                object type of the marker
            count_range : Tuple
                Min and max number of markers consumable ('cardinalities')
            marker_key : int
                Key of the marker
            """
            self.__related_activity = related_activity
            self.__object_type = object_type
            self.__count_range = count_range
            self.__marker_key = marker_key

        def __repr__(self):
            return f"(a={self.related_activity}, ot={self.object_type}, c={self.count_range}, k={self.marker_key})"

        def __str__(self):
            return self.__repr__()

        def __hash__(self):
            return hash(
                (
                    self.related_activity,
                    self.object_type,
                    self.min_count,
                    self.max_count,
                    self.marker_key,
                )
            )

        def __get_related_activity(self):
            return self.__related_activity

        def __get_object_type(self):
            return self.__object_type

        def __get_count_range(self):
            return self.__count_range

        def __get_min_count(self):
            return self.__count_range[0]

        def __get_max_count(self):
            return self.__count_range[1]

        def __get_marker_key(self):
            return self.__marker_key

        def __set_marker_key(self, marker_key: int):
            self.__marker_key = marker_key

        def __eq__(self, other):
            if isinstance(other, OCCausalNet.Marker):
                return (
                    self.related_activity == other.related_activity
                    and self.object_type == other.object_type
                    and self.min_count == other.min_count
                    and self.max_count == other.max_count
                    and self.marker_key == other.marker_key
                )
            return False

        related_activity = property(__get_related_activity)
        object_type = property(__get_object_type)
        count_range = property(__get_count_range)
        min_count = property(__get_min_count)
        max_count = property(__get_max_count)
        marker_key = property(__get_marker_key, __set_marker_key)

    class MarkerGroup(object):
        """
        Represents a group of markers. A group of markers semantically
        represents the AND gate of all markers in the group.
        """

        def __init__(
            self,
            markers: List["OCCausalNet.Marker"],
            support_count: int = float("inf"),
        ):
            """
            Constructor

            Parameters
            ----------
            markers : List[OCCausalNet.Marker]
                List of markers that comprise the group
            support_count : int
                Frequency of this marker group in the event log. May be used to
                filter infrequent marker groups.
                Default is inf.
            """
            self.__markers = markers
            self.__support_count = support_count

        def __repr__(self):
            return f"({self.markers}, count={self.support_count})"

        def __str__(self):
            return self.__repr__()

        def __eq__(self, other):
            if isinstance(other, OCCausalNet.MarkerGroup):
                return (
                    Counter(self.markers) == Counter(other.markers)
                    and self.support_count == other.support_count
                )
            return False

        def __hash__(self):
            return hash(
                (
                    frozenset(self.markers),
                    self.support_count,
                )
            )

        def __get_markers(self):
            return self.__markers

        def __get_support_count(self):
            return self.__support_count

        @cached_property
        def dict_representation(self):
            """
            Returns a dictionary representation of the marker group for
            efficient checking if the marker group can be bound with a
            given set of objects per related activity and object type.
            Is only computed once and cached.
            Is invalid if the marker group is changed after initialization.
            Assumes that the marker group is valid, i.e., there is at most one marker per
            related activity and object type.

            Returns
            -------
            defaultdict[str, defaultdict[str, tuple[int, int]]]
                Dictionary representation of the marker group, mapping
                related activities to objects types to min and max cardinalities.
            """
            result = defaultdict(lambda: defaultdict(lambda: (float("inf"), 0)))
            for marker in self.markers:
                related_activity = marker.related_activity
                object_type = marker.object_type
                result[related_activity][object_type] = (
                    marker.min_count,
                    marker.max_count if marker.max_count != -1 else float("inf"),
                )
            return result

        @cached_property
        def key_constraints(self):
            """
            Returns all tuples (related_activity, object_type, related_activity_2) that
            cannot share objects due to having the same key.
            Is only computed once and cached.
            Is invalid if the marker group is changed after initialization.

            Returns
            -------
            List[Tuple[str, str, str]]
                List of tuples (related_activity, object_type, related_activity_2)
                that cannot share the same marker key.
            """
            # group related activities by (marker_key, object_type)
            grouped = defaultdict(list)
            for marker in self.markers:
                grouped[(marker.marker_key, marker.object_type)].append(
                    marker.related_activity
                )

            # Generate constraints from groups with >= 2 elements
            constraints = []
            for (marker_key, object_type), related_activities in grouped.items():
                if len(related_activities) > 1:
                    for act1, act2 in combinations(related_activities, 2):
                        constraints.append((act1, object_type, act2))

            return constraints

        markers = property(__get_markers)
        support_count = property(__get_support_count)

    def __init__(
        self,
        dependency_graph: nx.MultiDiGraph,
        output_marker_groups: Dict[str, List["OCCausalNet.MarkerGroup"]],
        input_marker_groups: Dict[str, List["OCCausalNet.MarkerGroup"]],
        activity_count: Dict[str, int] = None,
        relative_occurrence_threshold: float = 0,
    ):
        """
        Constructor

        Parameters
        ----------
        dependency_graph : nx.MultiDiGraph
            Object-centric dependency graph
            Arc (a, object_type, a') must be encoded as dg[a][a'][object_type] = {"object_type": object_type}
        output_marker_groups : Dict[str, List[OCCausalNet.MarkerGroup]]
            Output marker groups per activity
        input_marker_groups : Dict[str, List[OCCausalNet.MarkerGroup]]
            Input marker groups per activity
        activity_count : Dict[str, int]
            Activity counts in the event log for filtering of infrequent marker groups.
        relative_occurrence_threshold : float
            Relative threshold for filtering infrequent marker groups. Range is [0,1].
            Default is 0, meaning no filtering.
        """
        self.__dependency_graph = dependency_graph
        self.__activities = list(dependency_graph._node.keys())
        if activity_count is None:
            activity_count = {act: 1 for act in self.activities}
        self.__edges = dependency_graph._succ
        self.__relative_occurrence_threshold = relative_occurrence_threshold
        self.__input_marker_groups, self.__output_marker_groups = filter4(
            input_marker_groups,
            output_marker_groups,
            self.__relative_occurrence_threshold,
            activity_count,
        )
        self.__object_types = {
            o.object_type
            for binds in self.__input_marker_groups.values()
            for bs in binds
            for o in bs.markers
        }
        self.__activity_count = activity_count

    def __repr__(self):
        # a OCCN is fully defined by its activities and marker groups
        ret = f"Activities: {self.activities}"
        for act in self.activities:
            img = (
                self.input_marker_groups[act] if act in self.input_marker_groups else []
            )
            ret += f"\nInput_marker_groups[{act}]: {img}\n"
            omg = (
                self.output_marker_groups[act]
                if act in self.output_marker_groups
                else []
            )
            ret += f"Output_marker_groups[{act}]: {omg}"
        return ret

    def __str__(self):
        return self.__repr__()

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        if isinstance(other, OCCausalNet):
            def normalize_edges(edges):
                normalized = Counter()
                for source, target_dict in edges.items():
                    for target, key_dict in target_dict.items():
                        for edge_key, attributes in key_dict.items():
                            object_type = (
                                attributes.get("object_type")
                                if isinstance(attributes, dict)
                                else None
                            )
                            normalized[(source, target, edge_key, object_type)] += 1
                return normalized

            return (
                set(self.activities) == set(other.activities)
                and normalize_edges(self.edges) == normalize_edges(other.edges)
                and all(
                    Counter(self.input_marker_groups.get(a, []))
                    == Counter(other.input_marker_groups.get(a, []))
                    for a in self.activities
                )
                and all(
                    Counter(self.output_marker_groups.get(a, []))
                    == Counter(other.output_marker_groups.get(a, []))
                    for a in self.activities
                )
                and set(self.object_types) == set(other.object_types)
                and all(
                    self.activity_count.get(a, 0) == other.activity_count.get(a, 0)
                    for a in self.activities
                )
                and self.relative_occurrence_threshold
                == other.relative_occurrence_threshold
            )
        return False

    def __get_dependency_graph(self):
        return self.__dependency_graph

    def __get_activities(self):
        return self.__activities

    def __get_edges(self):
        return self.__edges

    def __get_input_marker_groups(self):
        return self.__input_marker_groups

    def __get_output_marker_groups(self):
        return self.__output_marker_groups

    def __get_object_types(self):
        return self.__object_types

    def __get_activity_count(self):
        return self.__activity_count

    def __get_relative_occurrence_threshold(self):
        return self.__relative_occurrence_threshold

    dependency_graph = property(__get_dependency_graph)
    activities = property(__get_activities)
    edges = property(__get_edges)
    input_marker_groups = property(__get_input_marker_groups)
    output_marker_groups = property(__get_output_marker_groups)
    object_types = property(__get_object_types)
    activity_count = property(__get_activity_count)
    relative_occurrence_threshold = property(__get_relative_occurrence_threshold)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import os
import sqlite3
from datetime import datetime, timedelta
import pandas as pd
from enum import Enum
from typing import Optional, Dict, Any

from pm4py.util import exec_utils, pandas_utils


class Parameters(Enum):
    HISTORY_DB_PATH = "history_db_path"


# ---------------------------------------------------------------------------
# Helper: decode the transition core type stored in Chrome history
# ---------------------------------------------------------------------------
_transition_core_map = {
    0: "LINK",  # user clicked a link
    1: "TYPED",  # user typed the URL in the omnibox
    2: "AUTO_BOOKMARK",  # URL was bookmarked and launched
    3: "AUTO_SUBFRAME",  # sub‑frame navigation (auto)
    4: "MANUAL_SUBFRAME",  # sub‑frame navigation (manual)
    5: "GENERATED",  # generated by the browser (e.g. search)
    6: "START_PAGE",  # start‑up or home page
    7: "FORM_SUBMIT",  # form submission
    8: "RELOAD",  # explicit reload (F5, ⟳)
    9: "KEYWORD",  # keyword search in omnibox
    10: "KEYWORD_GENERATED",  # keyword search generated URL
}


def _decode_transition(value: int) -> str:
    """Return the human‑readable core transition type.

    Chrome encodes qualifiers in the upper bits; the lower 8 bits (0xFF) store
    the *core* transition type (enum content/public/common/page_transition_types.h).
    """
    core = value & 0xFF
    return _transition_core_map.get(core, f"UNKNOWN_{core}")


# ---------------------------------------------------------------------------
# Main extractor
# ---------------------------------------------------------------------------

def apply(parameters: Optional[Dict[Any, str]] = None) -> pd.DataFrame:
    """Extract a dataframe containing the navigation history of Google Chrome.

    **Event log semantics**
    -----------------------
    * **CASE ID** (``case:concept:name``) browser *profile* name (e.g. "Profile 1")
    * **ACTIVITY** (``concept:name``) full path of the visited page ("example.com/path"), *without* query parameters
    * **TIMESTAMP** (``time:timestamp``) local timestamp of the visit (start of the page view)

    **Additional attributes**
    --------------------------------
    * ``title`` page title as recorded by Chrome (empty if unavailable)
    * ``visit_duration`` seconds spent on the page (0 if unknown / Chrome <93)
    * ``transition`` human‑readable page transition (LINK, TYPED, …)
    * ``typed_count`` how many times the user manually typed this URL
    * ``visit_count`` how many *total* visits Chrome knows for this URL
    * ``referrer`` URL of the *from_visit* (previous page) if available

    Parameters
    ----------
    Parameters.HISTORY_DB_PATH : str, optional
        Path to the **folder** that contains the Chrome profile directories or to
        a single *History* SQLite file. Defaults to the standard Windows path of
        the current user (``%LOCALAPPDATA%\\Google\\Chrome\\User Data``).

    Returns
    -------
    pd.DataFrame
        PM4Py‑compatible event log where each row is a page visit.
    """

    if parameters is None:
        parameters = {}

    history_db_path = exec_utils.get_param_value(
        Parameters.HISTORY_DB_PATH,
        parameters,
        os.path.join(
            os.getenv("LOCALAPPDATA", ""), "Google", "Chrome", "User Data"
        ),
    )

    # ---------------------------------------------------------------------
    # Locate profile history files
    # ---------------------------------------------------------------------
    if os.path.isdir(history_db_path):
        profiles = [
            (os.path.join(history_db_path, prof, "History"), prof)
            for prof in os.listdir(history_db_path)
            if prof.startswith("Profile ") or prof == "Default"
        ]
    else:
        # user passed a single file instead of a directory
        profiles = [(history_db_path, "DEFAULT")]

    profiles = [p for p in profiles if os.path.exists(p[0])]

    events = []
    for history_file, profile_name in profiles:
        # open read‑only – Chrome keeps a *Write‑Ahead Log* that blocks writing
        conn = sqlite3.connect(f"file:{history_file}?mode=ro", uri=True)
        cur = conn.cursor()

        # Check if visit_duration column exists (Chrome 93+)
        cur.execute("PRAGMA table_info(visits)")
        visit_cols = [row[1] for row in cur.fetchall()]
        has_duration = "visit_duration" in visit_cols

        # Build query dynamically based on column availability
        query = f"""
            SELECT
                urls.url AS full_url,
                urls.title AS title,
                urls.typed_count AS typed_cnt,
                urls.visit_count AS visit_cnt,
                visits.visit_time AS visit_time,
                { 'visits.visit_duration' if has_duration else '0' } AS duration,
                visits.transition AS transition,
                visits.from_visit AS from_visit_id
            FROM visits
            JOIN urls ON visits.url = urls.id
        """
        cur.execute(query)
        rows = cur.fetchall()

        # Build *id → url* map so we can resolve from_visit later without a JOIN
        cur.execute("SELECT id, url FROM urls")
        id2url = dict(cur.fetchall())

        for (
            full_url,
            title,
            typed_cnt,
            visit_cnt,
            visit_time,
            duration,
            transition_val,
            from_visit_id,
        ) in rows:
            # ----------------------------------------------------------------
            # Convert Chromium timestamps (µs since 1601‑01‑01) → datetime
            # ----------------------------------------------------------------
            # Re‑implement the SQL the original script used so that we avoid
            # timezone surprises and keep everything in Python for clarity.
            visit_start = datetime(1601, 1, 1) + timedelta(microseconds=visit_time)
            visit_start = visit_start.astimezone()  # convert to local tz

            # Duration is stored in *microseconds* (uint32) – Chrome <93 omitted it
            if has_duration and duration:
                visit_duration = duration / 1_000_000.0
            else:
                visit_duration = 0.0

            referrer_url = id2url.get(from_visit_id, "")

            ev = {
                "case:concept:name": profile_name,
                "concept:name": full_url.split("//")[-1].split("?")[0].replace(",", ""),
                "complete_url": full_url,
                "domain": full_url.split("//")[-1].split("/")[0],
                "url_wo_parameters": full_url.split("//")[-1].split("?")[0],
                "time:timestamp": visit_start,
                # New attributes
                "title": title or "",
                "visit_duration": visit_duration,
                "transition": _decode_transition(transition_val),
                "typed_count": typed_cnt,
                "visit_count": visit_cnt,
                "referrer": referrer_url,
            }

            # Filter empty values for the mandatory XES attributes
            if ev["case:concept:name"].strip() and ev["concept:name"].strip():
                events.append(ev)

        cur.close()
        conn.close()

    # ---------------------------------------------------------------------
    # Build PM4Py dataframe (same post‑processing as original)
    # ---------------------------------------------------------------------
    df = pandas_utils.instantiate_dataframe(events)
    if not df.empty:
        df = pandas_utils.insert_index(df, "@@index", copy_dataframe=False, reset_index=False)
        df = df.sort_values(["time:timestamp", "@@index"])
        df["@@case_index"] = df.groupby("case:concept:name", sort=False).ngroup()
        df = df.sort_values(["@@case_index", "time:timestamp", "@@index"])
    return df



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.simulation.playout.oc_causal_net.variants import extensive
from pm4py.util import exec_utils
from enum import Enum
from pm4py.objects.oc_causal_net.obj import OCCausalNet
from typing import Optional, Dict, Any
from pm4py.objects.ocel.obj import OCEL  


class Variants(Enum):
    EXTENSIVE = extensive


DEFAULT_VARIANT = Variants.EXTENSIVE
VERSIONS = {Variants.EXTENSIVE}


def apply(occn: OCCausalNet, objects, parameters: Optional[Dict[Any, Any]] = None, variant=DEFAULT_VARIANT) -> OCEL:
    """
    Do the playout of an object-centric causal net generating an OCEL.

    Parameters
    -----------
    occn
        Object-centric causal net to play-out
    objects
        Dictionary mapping object types to object ids. These objects will be introduced by the start activities of the occn at the beginning of every binding sequence.
    parameters
        Parameters of the algorithm
    variant
        Variant of the algorithm to use:
            - Variants.EXTENSIVE: gets all the traces from the model. can be expensive
            
    Returns
    -----------
    OCEL
        Object-centric event log generated by the playout of the object-centric causal net
    """
    return exec_utils.get_variant(variant).apply(occn, objects, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any
from pm4py.util import exec_utils, pandas_utils
from enum import Enum
from pm4py.algo.connectors.util import mail as mail_utils
from pm4py.util.dt_parsing.variants import strpfromiso
import pandas as pd
from datetime import datetime, timedelta
import importlib.util
import traceback


class Parameters(Enum):
    EMAIL_USER = "email_user"
    CALENDAR_ID = "calendar_id"
    INCLUDE_REMINDERS = "include_reminders"


# Values used by Outlook that identify a cancelled meeting (olMeetingCanceled)
# See: https://learn.microsoft.com/office/vba/api/outlook.olmeetingstatus
OL_MEETING_CANCELED = 5


def _add_event(
    buffer: list,
    conversation_id: str,
    subject: str,
    timestamp: datetime,
    activity: str,
):
    """Helper to push a single event dict on *buffer* with minimal syntax."""

    buffer.append(
        {
            "case:concept:name": conversation_id,
            "case:subject": subject,
            "time:timestamp": timestamp,
            "concept:name": activity,
        }
    )


def apply(parameters: Optional[Dict[str, Any]] = None) -> pd.DataFrame:
    """
    Extracts the history of the calendar events (creation, update, start, end)
    in a Pandas dataframe from the local Outlook instance running on the current computer.

    CASE ID (case:concept:name) => identifier of the meeting
    ACTIVITY (concept:name) => one between: Meeting Created, Last Change of Meeting, Meeting Started, Meeting Completed,
    Meeting Cancelled, Meeting Reminder Fired
    TIMESTAMP (time:timestamp) => the timestamp of the event
    case:subject => the subject of the meeting

    Returns
    -------
    dataframe
        Pandas dataframe
    """
    if parameters is None:
        parameters = {}

    calendar_id = exec_utils.get_param_value(Parameters.CALENDAR_ID, parameters, 9)
    email_user = exec_utils.get_param_value(Parameters.EMAIL_USER, parameters, None)
    include_reminders = exec_utils.get_param_value(
        Parameters.INCLUDE_REMINDERS, parameters, True
    )

    # Connect to local Outlook instance / profile
    calendar = mail_utils.connect(email_user, calendar_id)

    # Optional progress bar if *tqdm* is on the system
    progress = None
    if importlib.util.find_spec("tqdm"):
        from tqdm.auto import tqdm

        progress = tqdm(total=len(calendar.Items), desc="extracting calendar items")

    events: list[Dict[str, Any]] = []

    for item in calendar.Items:
        try:
            conversation_id = str(item.ConversationID)
            subject = str(item.Subject)

            creation_time = strpfromiso.fix_naivety(
                datetime.fromtimestamp(item.CreationTime.timestamp())
            )
            last_mod_time = strpfromiso.fix_naivety(
                datetime.fromtimestamp(item.LastModificationTime.timestamp())
            )
            start_time = strpfromiso.fix_naivety(
                datetime.fromtimestamp(item.Start.timestamp())
            )
            end_time = strpfromiso.fix_naivety(
                datetime.fromtimestamp(item.Start.timestamp() + 60 * item.Duration)
            )

            # ── canonical 4 events ──────────────────────────────────────────────
            _add_event(events, conversation_id, subject, creation_time, "Meeting Created")

            if last_mod_time != creation_time:
                _add_event(
                    events,
                    conversation_id,
                    subject,
                    last_mod_time,
                    "Last Change of Meeting",
                )

            _add_event(events, conversation_id, subject, start_time, "Meeting Started")
            _add_event(events, conversation_id, subject, end_time, "Meeting Completed")

            # ── newly added events ─────────────────────────────────────────────
            # a) Meeting cancelled (status 5) – timestamped at *LastModificationTime*
            meeting_status = getattr(item, "MeetingStatus", None)
            if meeting_status == OL_MEETING_CANCELED:
                _add_event(events, conversation_id, subject, last_mod_time, "Meeting Cancelled")

            # b) Reminder – fire time = start_time − ReminderMinutesBeforeStart
            if include_reminders and getattr(item, "ReminderSet", False):
                minutes_before = getattr(item, "ReminderMinutesBeforeStart", None)
                if isinstance(minutes_before, int):
                    reminder_time = start_time - timedelta(minutes=minutes_before)
                    # Only add if the computed time is not in the future: the reminder has *already* triggered
                    if reminder_time <= datetime.now(tz=reminder_time.tzinfo):
                        _add_event(
                            events,
                            conversation_id,
                            subject,
                            reminder_time,
                            "Meeting Reminder Fired",
                        )

        except BaseException:
            traceback.print_exc()
            # swallow – keep going but still show failure in console

        if progress is not None:
            progress.update()

    if progress is not None:
        progress.close()

    # ── turn list → DataFrame & canonical ordering ───────────────────────────
    df = pandas_utils.instantiate_dataframe(events)
    df = pandas_utils.insert_index(df, "@@index", copy_dataframe=False, reset_index=False)
    df = df.sort_values(["time:timestamp", "@@index"])
    df["@@case_index"] = df.groupby("case:concept:name", sort=False).ngroup()
    df = df.sort_values(["@@case_index", "time:timestamp", "@@index"])

    return df



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from pm4py.util.lp import solver
from pm4py.util import constants
from pm4py.objects.petri_net.utils.petri_utils import remove_place
from typing import Tuple
import warnings
import importlib.util
from pm4py.objects.petri_net.obj import PetriNet, Marking


def apply_reduction(
    net: PetriNet, im: Marking, fm: Marking
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Apply the Murata reduction to an accepting Petri net, removing structurally redundant (implicit) places.

    This implementation follows the (Berthelot) implicit-place check by searching, via ILP,
    for a place-flow f = a_p * p - Σ a_q * q (a_p >= 1, a_q >= 0) such that:

      (1) f is a P-invariant (flow):  a_p*C(p,t) - Σ a_q*C(q,t) = 0    for all transitions t
          where C(p,t) = Post(p,t) - Pre(p,t)

      (2) a_p*Pre(p,t) - Σ a_q*Pre(q,t) <= k   for all transitions t
          with k = a_p*M0(p) - Σ a_q*M0(q),  and k >= 0

    If such a solution exists, the candidate place is implicit and can be removed without changing behavior.

    Parameters
    ---------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking

    Returns
    --------------
    net
        Reduced Petri net
    im
        Initial marking (unchanged)
    fm
        Final marking (unchanged)
    """
    places = sorted(list(net.places), key=lambda x: x.name)
    redundant = set()

    # Choose solver backend
    proposed_solver = solver.SCIPY
    if importlib.util.find_spec("pulp"):
        proposed_solver = solver.PULP
    else:
        if constants.SHOW_INTERNAL_WARNINGS:
            warnings.warn(
                "solution from scipy may be unstable. Please install PuLP (pip install pulp) for fully reliable results."
            )

    for place in places:
        # Skip already proven redundant places
        if place in redundant:
            continue

        # Skip places in the initial or final markings
        if place in im or place in fm:
            continue

        # Work on the "current" net as if previously found redundant places were removed:
        active_places = [p for p in places if p not in redundant]
        active_places = sorted(active_places, key=lambda x: x.name)
        place_index = {p: i for i, p in enumerate(active_places)}

        n_places = len(active_places)
        n_vars = n_places + 1  # last variable is k
        k_idx = n_vars - 1

        Aeq = []
        Aub = []
        beq = []
        bub = []

        # ---------------------------------------------------------------------
        # (A) k = a_p*M0(p) - Σ a_q*M0(q)
        #     => a_p*M0(p) - Σ a_q*M0(q) - k = 0
        # ---------------------------------------------------------------------
        eq = [0] * n_vars
        for p2 in active_places:
            m0 = im[p2] if p2 in im else 0
            if m0 == 0:
                continue
            if p2 == place:
                eq[place_index[p2]] += m0
            else:
                eq[place_index[p2]] -= m0
        eq[k_idx] = -1
        Aeq.append(eq)
        beq.append(0)

        # ---------------------------------------------------------------------
        # (B) Flow constraints: for all transitions t,
        #     a_p*C(p,t) - Σ a_q*C(q,t) = 0
        #     where C(p,t)=Post(p,t)-Pre(p,t)
        # ---------------------------------------------------------------------
        for trans in net.transitions:
            pre = {}
            post = {}

            for arc in trans.in_arcs:
                src = arc.source
                if src in place_index:
                    pre[src] = pre.get(src, 0) + arc.weight

            for arc in trans.out_arcs:
                tgt = arc.target
                if tgt in place_index:
                    post[tgt] = post.get(tgt, 0) + arc.weight

            eq = [0] * n_vars
            for p2 in active_places:
                c_val = post.get(p2, 0) - pre.get(p2, 0)
                if c_val == 0:
                    continue
                if p2 == place:
                    eq[place_index[p2]] += c_val
                else:
                    eq[place_index[p2]] -= c_val
            # k has coefficient 0 here
            Aeq.append(eq)
            beq.append(0)

        # ---------------------------------------------------------------------
        # (C) Implicitness inequalities: for all transitions t,
        #     a_p*Pre(p,t) - Σ a_q*Pre(q,t) - k <= 0
        # ---------------------------------------------------------------------
        for trans in net.transitions:
            ineq = [0] * n_vars
            for arc in trans.in_arcs:
                p2 = arc.source
                if p2 not in place_index:
                    continue
                if p2 == place:
                    ineq[place_index[p2]] += arc.weight
                else:
                    ineq[place_index[p2]] -= arc.weight
            ineq[k_idx] = -1
            Aub.append(ineq)
            bub.append(0)

        # ---------------------------------------------------------------------
        # (D) Variable lower bounds:
        #     a_p >= 1, a_q >= 0, k >= 0
        #     Implemented as -a <= b
        # ---------------------------------------------------------------------
        for p2 in active_places:
            ineq = [0] * n_vars
            ineq[place_index[p2]] = -1
            Aub.append(ineq)
            if p2 == place:
                bub.append(-1)  # -a_p <= -1  -> a_p >= 1
            else:
                bub.append(0)   # -a_q <= 0   -> a_q >= 0

        ineq = [0] * n_vars
        ineq[k_idx] = -1
        Aub.append(ineq)
        bub.append(0)  # -k <= 0 -> k >= 0

        # Objective (any feasible solution is enough; minimize sum to keep it small)
        c = [1] * n_vars
        integrality = [1] * n_vars

        xx = solver.apply(
            c,
            Aub,
            bub,
            Aeq,
            beq,
            variant=proposed_solver,
            parameters={"integrality": integrality},
        )

        if (hasattr(xx, "success") and xx.success) or (
            hasattr(xx, "sol_status") and xx.sol_status > -1
        ):
            redundant.add(place)

    # Remove redundant places (deterministic order)
    for pl in sorted(list(redundant), key=lambda x: x.name):
        net = remove_place(net, pl)

    return net, im, fm



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

import random
import sys
from pm4py.algo.simulation.playout.ocpn.variants.utils import feasible_traces_to_ocel
from pm4py.objects.ocpn.semantics import OCPetriNetSemantics
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocpn.obj import OCPetriNet, OCMarking
from pm4py.objects.ocel import constants
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union, List
from enum import Enum


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    MAX_BINDINGS_PER_ACTIVITY = "maxBindingsPerActivity"
    BRANCHING_FACTOR_TRANSITIONS = "branchingFactorTransitions"
    BRANCHING_FACTOR_BINDINGS = "branchingFactorBindings"
    OBJECTS_UNIQUE_PER_TRACE = "objects_unique_per_trace"
    RETURN_TRACES = "return_traces"
    EXISTS_TRACE = "exists_trace"
    OCPETRINET_SEMANTICS = "ocpetrinet_semantics"
    IS_FINAL_FUNC = "is_final_func"


FINAL_MARKER = "FINAL"


def apply(
    net: OCPetriNet,
    initial_marking: OCMarking,
    final_marking: OCMarking,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> OCEL:
    """
    Compute playout of an object-centric Petri net generating an OCEL (extensive search;
    any activity may only be executed a limited number of times as specified).
    Uses an optimized Memoized DFS Graph Population algorithm.

    Parameters
    -----------
    net
        Object-centric Petri net to play-out
    initial_marking
        Initial marking of the object-centric Petri net
    final_marking
        Final marking of the object-centric Petri net
    parameters
        Parameters of the algorithm:
            Parameters.MAX_BINDINGS_PER_ACTIVITY -> Maximum bindings per activity (mandatory)
            Parameters.EXISTS_TRACE -> If True, return a boolean indicating if at least one trace exists instead of an OCEL
            Parameters.OBJECTS_UNIQUE_PER_TRACE: If True, objects in the resulting OCEL are make unique per trace (default: False)
            Parameters.RETURN_TRACES -> If True, return traces instead of OCEL
            Parameters.BRANCHING_FACTOR_TRANSITIONS -> Maximum number of transitions to explore from a single
                state (default: sys.maxsize, i.e., no limit). If set to a float, it will be stochastically rounded to an integer for every state.
            Parameters.BRANCHING_FACTOR_BINDINGS -> Maximum number of bindings to explore for a single transition
                (default: sys.maxsize, i.e., no limit). If set to a float, it will be stochastically rounded to an integer for every state.
            Parameters.OCPETRINET_SEMANTICS -> Object-centric Petri net semantics
            Parameters.IS_FINAL_FUNC -> Function that given a marking and the final marking returns whether the final marking is reached.
                (default: marking == final_marking)
    """
    if parameters is None:
        parameters = {}

    return_traces = exec_utils.get_param_value(
        Parameters.RETURN_TRACES, parameters, False
    )
    exists_trace = exec_utils.get_param_value(
        Parameters.EXISTS_TRACE, parameters, False
    )
    if Parameters.MAX_BINDINGS_PER_ACTIVITY not in parameters:
        raise ValueError(
            "Parameter MAX_BINDINGS_PER_ACTIVITY must be specified for the extensive playout. This parameter limits the maximum number of times an activity may be executed."
        )
    max_bindings_per_activity = exec_utils.get_param_value(
        Parameters.MAX_BINDINGS_PER_ACTIVITY, parameters, None
    )
    # How many enabled transitions to explore from a single state
    # deactivated by default
    bf_trans = exec_utils.get_param_value(
        Parameters.BRANCHING_FACTOR_TRANSITIONS, parameters, sys.maxsize
    )
    # How many bindings to explore for a single transition
    # deactivated by default
    bf_binds = exec_utils.get_param_value(
        Parameters.BRANCHING_FACTOR_BINDINGS, parameters, sys.maxsize
    )
    semantics = exec_utils.get_param_value(
        Parameters.OCPETRINET_SEMANTICS,
        parameters,
        OCPetriNetSemantics(),
    )
    is_final = exec_utils.get_param_value(
        Parameters.IS_FINAL_FUNC,
        parameters,
        _is_final
    )

    # Save transitions as ids for memory efficiency; create lookup table for conversion
    all_transitions = sorted(list(net.transitions), key=lambda t: t.name)
    transition_to_idx = {t: i for i, t in enumerate(all_transitions)}

    # State key: (marking, number of firings per transition)
    # Transition firing counts; Position corresponds to transitions_to_idx table
    initial_transition_counts = (0,) * len(all_transitions)

    initial_state_key = (initial_marking, initial_transition_counts)

    # Memoization cache: Dict[state_key, Union[Set[Tuple[binding, next_key]], str]]
    # where the state_key is a tuple of (marking, transition_counts) and the
    # value is either FINAL_MARKER if the state is final,
    # or a set of tuples (binding, next_state_key) with all possible bindings
    # in the given state key and the state key they lead to
    memo = {}

    # == Phase 1: Memoization DFS Graph Population ==
    trace_exists = _populate_memo_graph(
        initial_state_key,
        net,
        final_marking,
        semantics,
        transition_to_idx,
        max_bindings_per_activity,
        bf_trans,
        bf_binds,
        exists_trace,
        is_final,
        memo,
    )
    
    # If we are only interested in whether a trace exists, return immediately
    if exists_trace:
        return trace_exists
    

    # == Phase 2: Reconstruct traces from memo ==
    feasible_traces_iter = _reconstruct_traces(
        initial_state_key, memo, transition_to_idx, all_transitions
    )

    # == Phase 3: Return data in desired format ==
    # We did not save the object types of objects in the events, so we need to reconstruct them from the initial marking
    # Maps object ids to their types
    id_to_obj_type = dict()
    # Every object can be found in the initial marking
    for p, obj_ids in initial_marking.items():
        ot = p.object_type
        for obj_id in obj_ids:
            id_to_obj_type[obj_id] = ot
    
    if return_traces:
        # Inverse the transition_to_idx mapping to get transition labels
        idx_to_transition = {v: k for k, v in transition_to_idx.items()}
        return (list(feasible_traces_iter), idx_to_transition, id_to_obj_type)
    else:
        return feasible_traces_to_ocel(
            feasible_traces_iter, all_transitions, id_to_obj_type, parameters
        )


def _populate_memo_graph(
    state_key: tuple,
    net: OCPetriNet,
    final_marking: OCMarking,
    semantics,
    t_to_idx: dict,
    max_bindings: int,
    bf_trans: float,
    bf_binds: float,
    exists_trace: bool,
    is_final,
    memo: dict,
) -> bool:
    """
    Recursively explores the state space to build a compact, memoized graph of all valid traces.

    This function performs a depth-first search from a given state_key. It populates a memoization
    cache (`memo`). For each state (defined by the state_key), it stores the set of "next steps" (as tuples of
    (transition_index, binding, next_state_key)) that lie on a path to the final marking,
    where binding is a set of object IDs that were bound to the transition.

    This approach avoids duplicate computation of two different traces leading to the same state key.

    Parameters
    -----------
    state_key
        A tuple `(OCMarking, tuple)` representing the current state of the search.
    net
        The object-centric Petri net being explored.
    final_marking
        The target OCMarking that signifies the end of a successful trace.
    semantics
        The OCPN semantics object used for enabling and firing transitions.
    t_to_idx
        A lookup dictionary mapping transition objects to their integer indices.
    max_bindings
        The maximum number of times any single transition is allowed to fire.
    bf_trans
        The max branching factor for transitions, limiting how many enabled transitions to explore.
        If set to a float, it will be stochastically rounded to an integer.
    bf_binds
        The max branching factor for bindings, limiting how many bindings to explore for each transition.
        If set to a float, it will be stochastically rounded to an integer.
    exists_trace
        If `True`, the function will return a boolean indicating if at least one trace exists that
        leads to the final marking, rather than populating the memo with all valid traces.
    is_final
        A function to determine if a marking is the final marking.
    memo
        The memoization cache, a dictionary that is modified in place by the function.
        It maps state_keys to the set of valid next steps or a special marker.

    Returns
    -----------
    bool
        Returns `True` if the final_marking is reachable from the given state_key,
        otherwise returns `False`.
    """
    if state_key in memo:
        # a deadlock is indicated by an empty set in the memo.
        # an entry that is not empty indicates that the final marking is reachable
        return bool(memo[state_key])

    # state_key has not been explored yet, so we explore it
    marking, transition_counts = state_key
    if is_final(marking, final_marking):
        # signal that the final marking is reached
        memo[state_key] = FINAL_MARKER
        return True

    next_steps = set()
    enabled_transitions = semantics.enabled_transitions(net, marking)
    # Stochastically round bf_trans to an integer
    rounded_bf_trans = int(bf_trans) + (1 if random.random() < (bf_trans % 1) else 0)
    # select a random subset if branching factor is limited
    transitions_to_explore = random.sample(
        list(enabled_transitions), k=min(rounded_bf_trans, len(enabled_transitions))
    )

    # explore all successor states by firing enabled transitions
    for t in transitions_to_explore:
        transition_idx = t_to_idx[t]
        if transition_counts[transition_idx] >= max_bindings:
            continue

        # Create new transition counts
        new_transition_counts = list(transition_counts)
        new_transition_counts[transition_idx] += 1
        new_counts_tuple = tuple(new_transition_counts)

        # Get all possible bindings for the transition
        bindings = list(semantics.get_possible_bindings(net, t, marking))

        # Stochastically round bf_binds to an integer
        rounded_bf_binds = int(bf_binds) + (
            1 if random.random() < (bf_binds % 1) else 0
        )
        # select a random subset of bindings if branching factor is limited
        bindings_to_explore = random.sample(
            bindings, k=min(rounded_bf_binds, len(bindings))
        )

        for binding in bindings_to_explore:
            new_marking = semantics.fire(net, t, marking, binding)
            new_state_key = (new_marking, new_counts_tuple)
            object_ids = [oi for objs in binding.values() for oi in objs]

            # Recursively explore the next state. Only add to memo it if it leads to the final state
            if _populate_memo_graph(
                new_state_key,
                net,
                final_marking,
                semantics,
                t_to_idx,
                max_bindings,
                bf_trans,
                bf_binds,
                exists_trace,
                is_final,
                memo,
            ):
                next_steps.add((transition_idx, frozenset(object_ids), new_state_key))
                if exists_trace:
                    # If we are only interested in whether a trace exists, we can return early
                    return True

    # Add all next steps to the memoization cache
    # If state_key is a deadlock, it will be an empty set
    memo[state_key] = next_steps
    return bool(next_steps)


def _reconstruct_traces(
    state_key: tuple, memo: dict, t_to_idx: dict, all_transitions: List
) -> iter:
    """
    Recursively reconstructs all valid traces by traversing the pre-computed graph in the memo cache.

    This function is a generator that operates on the populated `memo` dictionary created by
    `_populate_memo_graph`. It starts from a given state_key and recursively follows all valid
    "next steps" stored in the cache. When it reaches a path's end (marked by the FINAL_MARKER),
    it yields a complete trace.

    Parameters
    -----------
    state_key
        The tuple `(OCMarking, tuple)` representing the current state from which to reconstruct paths.
    memo
        The pre-populated memoization cache containing the valid path graph from the
        `_populate_memo_graph` function.
    t_to_idx
        A lookup dictionary mapping transition objects to their integer indices.
    all_transitions
        A ordered list of all transition objects in the Petri net.

    Yields
    -----------
    tuple
        A single, complete trace. Each yielded trace is a tuple of event tuples, where each
        event is `(transition_idx, frozenset(object_ids))`.
    """
    next_steps = memo.get(state_key)

    if next_steps == FINAL_MARKER:
        # If we reached the final marking, yield an empty trace
        yield ()
        return

    if not next_steps:
        # This should not happen unless the initial marking is a deadlock
        return

    for t_idx, object_ids, next_state_key in next_steps:
        current_event = (t_idx, object_ids)

        # Recursively yield all traces from the next state
        for tail in _reconstruct_traces(
            next_state_key, memo, t_to_idx, all_transitions
        ):
            # Prepend the current event
            yield (current_event,) + tail


def _is_final(marking, final_marking):
    return marking == final_marking


"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from pm4py.objects.oc_causal_net.obj import OCCausalNet
from pm4py.objects.ocpn.obj import OCPetriNet
import networkx as nx


AUX_ACTIVITY_PREFIX = "_silent_aux_"
POST_AUX_ACTIVITY_PREFIX = AUX_ACTIVITY_PREFIX + "out_"
PRE_AUX_ACTIVITY_PREFIX = AUX_ACTIVITY_PREFIX + "in_"


def apply(ocpn: OCPetriNet, parameters=None) -> OCCausalNet:
    """
    Convets an Object-centric Petri Net to an Object-centric Causal Net.

    Parameters
    ----------
    ocpn: OCPetriNet
        The Object-centric Petri Net to be converted.
    parameters: dict, optional
        Additional parameters for the conversion (not used in this implementation).

    Returns
    ----------
    OCCausalNet: OCCausalNet
        The resulting Object-centric Causal Net.
    """
    # assert no place start with AUX_ACTIVITY_PREFIX to be able to identify auxiliary places
    for p in ocpn.places:
        if p.name.startswith(AUX_ACTIVITY_PREFIX):
            raise ValueError(
                f"Place {p.name} starts with the reserved prefix '{AUX_ACTIVITY_PREFIX}'."
            )

    places = ocpn.places
    transitions = ocpn.transitions
    object_types = ocpn.object_types

    dependencies = dict()
    input_marker_groups = dict()
    output_marker_groups = dict()

    # get multi-object types per transition
    multi_ots = get_multi_object_types(ocpn)

    # create dependencies and marker groups for transitions
    transition_dependencies_marker_groups(
        transitions,
        multi_ots,
        dependencies,
        input_marker_groups,
        output_marker_groups,
    )

    # create dependencies and marker groups for places
    place_dependencies_marker_groups(
        places,
        multi_ots,
        dependencies,
        input_marker_groups,
        output_marker_groups,
    )

    # handle start and end places
    start_places = {
        ot: [p for p in ocpn.initial_marking.places if p.object_type == ot]
        for ot in ocpn.object_types
    }
    end_places = {
        ot: [p for p in ocpn.final_marking.places if p.object_type == ot]
        for ot in ocpn.object_types
    }

    # add START and END activities and their dependencies and marker groups
    start_end_act_dependencies_marker_groups(
        object_types,
        dependencies,
        input_marker_groups,
        output_marker_groups,
        start_places,
        end_places,
    )

    # add START and END activities to marker groups of start and end places
    add_start_end_act_markers(
        object_types,
        input_marker_groups,
        output_marker_groups,
        start_places,
        end_places,
    )

    # create the occn
    occn = OCCausalNet(
        dependency_graph=nx.MultiDiGraph(dependencies),
        output_marker_groups=output_marker_groups,
        input_marker_groups=input_marker_groups,
    )

    return occn


def get_multi_object_types(ocpn: OCPetriNet) -> dict:
    """
    Returns a dictionary mapping transition names to two dictionaries of multi-input/multi-output
    object types
    mapping object types to True if they are variable and False otherwise.
    A multi-input/output object type is one that has at least two incoming/outgoing arcs
    to/from the transition.

    Parameters
    ----------
    ocpn: OCPetriNet
        The Object-centric Petri Net to analyze.

    Returns
    -------
    dict: Dictionary with transition names as keys a dictionary mapping "input" and "output"
          to dictionaries of multi-input/multi-output object types to True/False.
    """
    multi_ots = {}
    for t in ocpn.transitions:
        multi_ots[t.name] = {"input": dict(), "output": dict()}
        in_ots = [arc.object_type for arc in t.in_arcs]
        multi_ots[t.name]["output"] = dict()
        out_ots = [arc.object_type for arc in t.out_arcs]

        for ot in set(in_ots):
            if in_ots.count(ot) > 1:
                multi_ots[t.name]["input"][ot] = any(
                    arc.is_variable for arc in t.in_arcs if arc.object_type == ot
                )

        for ot in set(out_ots):
            if out_ots.count(ot) > 1:
                multi_ots[t.name]["output"][ot] = any(
                    arc.is_variable for arc in t.out_arcs if arc.object_type == ot
                )

    return multi_ots


def get_aux_place_name(transition_name, object_type, is_input_aux):
    """
    Returns the name of the auxiliary place for a given transition and object type.

    Parameters
    ----------
    transition_name: str
        The name of the transition.
    object_type: str
        The object type associated with the auxiliary place.
    is_input_aux: bool
        Whether the auxiliary place is an input auxiliary place (True) or output auxiliary place (False).

    Returns
    -------
    str: The name of the auxiliary place.
    """
    prefix = PRE_AUX_ACTIVITY_PREFIX if is_input_aux else POST_AUX_ACTIVITY_PREFIX
    return f"{prefix}{transition_name}_{object_type}"


def transition_dependencies_marker_groups(
    transitions,
    multi_ots,
    dependencies,
    input_marker_groups,
    output_marker_groups,
):
    """
    Creates dependencies and marker groups for transitions in the Object-centric Petri Net.
    Each transition has one input marker group featuring a marker for each input arc,
    and one output marker group featuring a marker for each output arc.

    Parameters
    ----------
    transitions: list
        List of transitions in the Object-centric Petri Net.
    multi_ots: dict
        Dictionary mapping transition names to multi-object types.
    dependencies: dict
        Dictionary to store dependencies between activities.
    input_marker_groups: dict
        Dictionary to store input marker groups.
    output_marker_groups: dict
        Dictionary to store output marker groups.
    """
    for t in transitions:
        # dependencies
        dependencies[t.name] = dict()  # add as activity

        # get multi object types for this transition
        multi_input = multi_ots.get(t.name, dict()).get("input", dict())
        multi_output = multi_ots.get(t.name, dict()).get("output", dict())

        for arc in t.in_arcs:
            if arc.object_type in multi_input:
                # aux. place goes in between source place and transition
                aux_place_name = get_aux_place_name(
                    t.name, arc.object_type, is_input_aux=True
                )
                add_dependency(
                    dependencies, arc.source, aux_place_name, arc.object_type
                )
                add_dependency(dependencies, aux_place_name, t, arc.object_type)
            else:
                add_dependency(dependencies, arc.source, t, arc.object_type)
        for arc in t.out_arcs:
            if arc.object_type in multi_output:
                # aux. place goes in between transition and target place
                aux_place_name = get_aux_place_name(
                    t.name, arc.object_type, is_input_aux=False
                )
                add_dependency(dependencies, t, aux_place_name, arc.object_type)
                add_dependency(
                    dependencies, aux_place_name, arc.target, arc.object_type
                )
            else:
                add_dependency(dependencies, t, arc.target, arc.object_type)

        # markers from input aux activities
        input_aux_markers = [
            OCCausalNet.Marker(
                related_activity=(
                    get_aux_place_name(t.name, object_type, is_input_aux=True)
                ),
                object_type=object_type,
                count_range=(0, float("inf")) if multi_input[object_type] else (1, 1),
                marker_key=get_next_key(),
            )
            for object_type in multi_input
        ]

        # single input marker group
        input_marker_groups[t.name] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=arc.source.name,
                        object_type=arc.object_type,
                        count_range=(0, float("inf")) if arc.is_variable else (1, 1),
                        marker_key=get_next_key(),
                    )
                    for arc in t.in_arcs
                    if arc.object_type not in multi_input
                ]
                + input_aux_markers
            )
        ]

        # markers to output aux activities
        output_aux_markers = [
            OCCausalNet.Marker(
                related_activity=(
                    get_aux_place_name(t.name, object_type, is_input_aux=False)
                ),
                object_type=object_type,
                count_range=(0, float("inf")) if multi_output[object_type] else (1, 1),
                marker_key=get_next_key(),
            )
            for object_type in multi_output
        ]

        # single output marker group
        output_marker_groups[t.name] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=(arc.target.name),
                        object_type=arc.object_type,
                        count_range=(0, float("inf")) if arc.is_variable else (1, 1),
                        marker_key=get_next_key(),
                    )
                    for arc in t.out_arcs
                    if arc.object_type not in multi_output
                ]
                + output_aux_markers
            )
        ]

        # add marker groups for input auxiliary places
        for ot in multi_input:
            aux_place_name = get_aux_place_name(t.name, ot, is_input_aux=True)
            output_marker_groups[aux_place_name] = [
                # single marker group with one marker for the transition
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=t.name,
                            object_type=ot,
                            count_range=(1, 1),
                            marker_key=get_next_key(),
                        )
                    ]
                )
            ]
            input_marker_groups[aux_place_name] = [
                # marker for every predecessor of t with this object type
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=arc.source.name,
                            object_type=ot,
                            count_range=(1, 1),
                            marker_key=get_next_key(),
                        )
                        for arc in t.in_arcs
                        if arc.object_type == ot
                    ]
                )
            ]

        # add marker groups for output auxiliary places
        for ot in multi_output:
            aux_place_name = get_aux_place_name(t.name, ot, is_input_aux=False)
            input_marker_groups[aux_place_name] = [
                # single marker group with one marker for the transition
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=t.name,
                            object_type=ot,
                            count_range=(1, 1),
                            marker_key=get_next_key(),
                        )
                    ]
                )
            ]
            output_marker_groups[aux_place_name] = [
                # marker for every successor of t with this object type
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=arc.target.name,
                            object_type=ot,
                            count_range=(1, 1),
                            marker_key=get_next_key(),
                        )
                        for arc in t.out_arcs
                        if arc.object_type == ot
                    ]
                )
            ]


def place_dependencies_marker_groups(
    places, multi_ots, dependencies, input_marker_groups, output_marker_groups
):
    """
    Creates dependencies and marker groups for places in the Object-centric Petri Net.
    Each place has one input marker group per input arc having one marker each,
    and one output marker group per output arc having one marker each.

    Parameters
    ----------
    places: list
        List of places in the Object-centric Petri Net.
    multi_ots: dict
        Dictionary mapping transition names to multi-object types.
    dependencies: dict
        Dictionary to store dependencies between activities.
    input_marker_groups: dict
        Dictionary to store input marker groups.
    output_marker_groups: dict
        Dictionary to store output marker groups.
    """

    def is_predecessor_transition_with_multi_variant(arc):
        """
        Checks if the source of the arc is a transition with and the object type
        of the arc is a multi-output object type for the given transition.
        """
        return isinstance(
            arc.source, OCPetriNet.Transition
        ) and arc.object_type in multi_ots.get(arc.source.name, dict()).get(
            "output", dict()
        )

    def is_successor_transition_with_multi_variant(arc):
        """
        Checks if the target of the arc is a transition with and the object type
        of the arc is a multi-input object type for the given transition.
        """
        return isinstance(
            arc.target, OCPetriNet.Transition
        ) and arc.object_type in multi_ots.get(arc.target.name, dict()).get(
            "input", dict()
        )

    for p in places:
        # dependencies
        if p.name not in dependencies:
            dependencies[p.name] = dict()  # add as activity
        for arc in p.in_arcs:
            if not is_predecessor_transition_with_multi_variant(arc):
                add_dependency(dependencies, arc.source, p, arc.object_type)
            else:
                pass  # dependency for auxiliary places was already added
        for arc in p.out_arcs:
            if not is_successor_transition_with_multi_variant(arc):
                add_dependency(dependencies, p, arc.target, arc.object_type)
            else:
                pass  # dependency for auxiliary places was already added

        # one-element marker group per arc
        input_marker_groups[p.name] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=(
                            arc.source.name
                            if not is_predecessor_transition_with_multi_variant(arc)
                            else get_aux_place_name(
                                arc.source.name, arc.object_type, is_input_aux=False
                            )
                        ),  # connect to aux place instead if multi-variant ot
                        object_type=arc.object_type,
                        count_range=(1, float("inf")),
                        marker_key=get_next_key(),
                    )
                ]
            )
            for arc in p.in_arcs
        ]

        output_marker_groups[p.name] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=(
                            arc.target.name
                            if not is_successor_transition_with_multi_variant(arc)
                            else get_aux_place_name(
                                arc.target.name, arc.object_type, is_input_aux=True
                            )
                        ),
                        object_type=arc.object_type,
                        count_range=(1, float("inf")),
                        marker_key=get_next_key(),
                    )
                ]
            )
            for arc in p.out_arcs
        ]


def start_end_act_dependencies_marker_groups(
    object_types,
    dependencies,
    input_marker_groups,
    output_marker_groups,
    start_places,
    end_places,
):
    """
    Adds a START and END activity for each object type.
    Creates dependencies and marker groups for those new activities.
    Each START / END activity has one input / output marker group consisting
    of markers for every start / end place.

    Parameters
    ----------
    object_types: list
        List of object types in the Object-centric Petri Net.
    dependencies: dict
        Dictionary to store dependencies between activities.
    input_marker_groups: dict
        Dictionary to store input marker groups.
    output_marker_groups: dict
        Dictionary to store output marker groups.
    start_places: dict
        Dictionary mapping object types to their start places.
    end_places: dict
        Dictionary mapping object types to their end places.
    """
    for ot in object_types:
        # START places
        dependencies[f"START_{ot}"] = dict()
        for p in start_places[ot]:
            add_dependency(dependencies, f"START_{ot}", p, ot)

        # one output marker group
        output_marker_groups[f"START_{ot}"] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=p.name,
                        object_type=ot,
                        count_range=(1, float("inf")),
                        marker_key=get_next_key(),
                    )
                    for p in start_places[ot]
                ]
            )
        ]

        # END places
        dependencies[f"END_{ot}"] = dict()
        for p in end_places[ot]:
            add_dependency(dependencies, p, f"END_{ot}", ot)

        # one input marker group
        input_marker_groups[f"END_{ot}"] = [
            OCCausalNet.MarkerGroup(
                [
                    OCCausalNet.Marker(
                        related_activity=p.name,
                        object_type=ot,
                        count_range=(1, float("inf")),
                        marker_key=get_next_key(),
                    )
                    for p in end_places[ot]
                ]
            )
        ]


def add_start_end_act_markers(
    object_types, input_marker_groups, output_marker_groups, start_places, end_places
):
    """
    Adds markers for the START and END activities to the start / end places.
    These markers are the same as for any other predecessor / successor activity.

    Parameters
    ----------
    object_types: list
        List of object types in the Object-centric Petri Net.
    input_marker_groups: dict
        Dictionary to store input marker groups.
    output_marker_groups: dict
        Dictionary to store output marker groups.
    start_places: dict
        Dictionary mapping object types to their start places.
    end_places: dict
        Dictionary mapping object types to their end places.
    """
    for ot in object_types:
        for p in start_places[ot]:
            # add new marker group
            if p.name not in input_marker_groups:
                input_marker_groups[p.name] = []
            input_marker_groups[p.name].append(
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=f"START_{ot}",
                            object_type=ot,
                            count_range=(1, float("inf")),
                            marker_key=get_next_key(),
                        )
                    ]
                )
            )
        for p in end_places[ot]:
            # add new marker group
            if p.name not in output_marker_groups:
                output_marker_groups[p.name] = []
            output_marker_groups[p.name].append(
                OCCausalNet.MarkerGroup(
                    [
                        OCCausalNet.Marker(
                            related_activity=f"END_{ot}",
                            object_type=ot,
                            count_range=(1, float("inf")),
                            marker_key=get_next_key(),
                        )
                    ]
                )
            )


def add_dependency(dependencies: dict, source, target, object_type):
    """
    Adds a dependency to the dependencies dictionary. Ignores duplicate dependencies.

    Parameters
    ----------
    dependencies: dict
        The dictionary to which the dependency will be added.
    source: str or OCPetriNet.Place or OCPetriNet.Transition
        The source of the dependency arc.
    target str or OCPetriNet.Place or OCPetriNet.Transition
        The target of the dependency arc.
    object_type
        The object type associated with the dependency.
    """
    source_label = source if isinstance(source, str) else source.name
    target_label = target if isinstance(target, str) else target.name
    if source_label not in dependencies:
        dependencies[source_label] = dict()
    if target_label not in dependencies[source_label]:
        dependencies[source_label][target_label] = dict()
    if object_type not in dependencies[source_label][target_label]:
        dependencies[source_label][target_label][object_type] = {
            "object_type": object_type
        }


def get_next_key():
    """
    Will return a unique key for each call, starting from 0.
    """
    # initialize on first call
    if not hasattr(get_next_key, "counter"):
        get_next_key.counter = 0
    current = get_next_key.counter
    get_next_key.counter += 1
    return current



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
# Author: Maximilian Josef Frank (https://orcid.org/0000-0002-0714-7748)

from typing import Dict, FrozenSet, List

class GeneticMatrix:
    def __init__(
        self,
        input_map: Dict[str, List[FrozenSet]],
        output_map: Dict[str, List[FrozenSet]],
        transitions: List[str],
    ):
        """
        Initialize a Genetic matrix

        Reference paper:
        Maximilian Josef Frank. "Optimising and Implementing the Genetic Miner in PM4Py" (2026).

        Parameters
        -------------
        input_map
            Input map for each node
        output_map
            Output map for each node
        transitions
            List of transitions
        """
        self.input_map = input_map
        self.output_map = output_map
        self.transitions = transitions

    def __repr__(self):
        return str({
            "I": self.input_map,
            "O": self.output_map,
            "T": self.transitions
        })

    def __str__(self):
        return self.__repr__()
